from typing import List
import dspy

class TextRelations(dspy.Signature):
  """Extract subject-predicate-object triples from the source text. Subject and object must be entities. Entities provided were previously extracted from the same source text.
  This is for an extraction task, please be thorough, accurate, and faithful to the reference text."""
  
  source_text: str = dspy.InputField()
  entities: str = dspy.InputField()
  relations: list[tuple[str, str, str]] = dspy.OutputField()

class ConversationRelations(dspy.Signature):
  """Extract subject-predicate-object triples from the conversation, including:
  1. Relations between concepts discussed
  2. Relations between speakers and concepts (e.g. user asks about X)
  3. Relations between speakers (e.g. assistant responds to user)
  Subject and object must be entities. Entities provided were previously extracted from the same source text.
  This is for an extraction task, please be thorough, accurate, and faithful to the reference text.
  """
  
  source_text: str = dspy.InputField()
  entities: str = dspy.InputField()
  relations: list[tuple[str, str, str]] = dspy.OutputField()

def get_relations(dspy: dspy.dspy, input_data: str, entities: list[str], is_conversation: bool = False) -> List[str]:
  if is_conversation:
    extract = dspy.Predict(ConversationRelations)
  else:
    extract = dspy.Predict(TextRelations)
    
  result = extract(source_text=input_data, entities=entities)
  filtered_relations = [
    (s, p, o) for s, p, o in result.relations 
    if s in entities and o in entities
  ]
  return filtered_relations


from src.kg_gen import KGGen
import os
from dotenv import load_dotenv

if __name__ == "__main__":
  # Load environment variables
  load_dotenv()
  
  # Example usage
  kg = KGGen()
  
  # Load fresh wiki content
  with open('tests/data/fresh_wiki_article.md', 'r', encoding='utf-8') as f:
    text = f.read()
  
  # Generate graph from wiki text
  graph = kg.generate(
    input_data=text,
    model="openai/gpt-4o",
    api_key=os.getenv("OPENAI_API_KEY")
  )
  print(graph)



from src.kg_gen import KGGen
import os
from dotenv import load_dotenv

if __name__ == "__main__":
  # Load environment variables
  load_dotenv()
  
  # Example usage
  kg = KGGen()
  
  # Test text input
  text = "Linda is Josh's mother. Ben is Josh's brother. Andrew is Josh's father. Judy is Andrew's sister. Josh is Judy's nephew. Judy is Josh's aunt."
  # Test with different models and their corresponding API keys
  model_configs = [
    {
      "model": "openai/gpt-4o",
      "api_key": os.getenv("OPENAI_API_KEY")
    },
    {
      "model": "anthropic/claude-3-5-sonnet-20240620", 
      "api_key": os.getenv("ANTHROPIC_API_KEY")
    },
    {
      "model": "gemini/gemini-pro",
      "api_key": os.getenv("GEMINI_API_KEY") 
    }
  ]
  
  for config in model_configs:
    print(f"\nTesting with model: {config['model']}")
    try:
      graph = kg.generate(
        input_data=text,
        model=config['model'],
        api_key=config['api_key']
      )
      print(graph)
    except Exception as e:
      print(f"Error with {config['model']}: {str(e)}")



from typing import Union, List, Dict, Tuple, Optional
from pydantic import BaseModel, model_validator, Field
from .steps._1_get_entities import get_entities
from .steps._2_get_relations import get_relations
import dspy
import json
import os

# ~~~ DATA STRUCTURES ~~~

class Graph(BaseModel):
  entities: set[str] = Field(..., description="All entities including additional ones from response")
  edges: set[str] = Field(..., description="All edges")
  relations: set[Tuple[str, str, str]] = Field(..., description="List of (subject, predicate, object) triples")

  @model_validator(mode='after')
  def validate_entities_in_relations(self) -> 'Graph':
    entities = set(self.entities)
    edges = set(self.edges)
    for subj, pred, obj in self.relations:
      if subj not in entities:
        raise ValueError(f"Relation subject '{subj}' not in entities")
      if obj not in entities:
        raise ValueError(f"Relation object '{obj}' not in entities")
      if pred not in edges:
        raise ValueError(f"Relation pred '{pred}' not edges")
    return self

  
# ~~~ KGGEN ~~~

class KGGen:
  def __init__(self):
    """Initialize KGGen"""
    self.dspy = dspy
    
  def generate(
    self,
    input_data: Union[str, List[Dict]],
    model: str,
    api_key: str = None,
    context: Optional[str] = None,
    example_relations: Optional[Union[
      List[Tuple[str, str, str]],
      List[Tuple[Tuple[str, str], str, Tuple[str, str]]]
    ]] = None,
    chunk_size: Optional[int] = None,
    node_labels: Optional[List[str]] = None,
    edge_labels: Optional[List[str]] = None,
    ontology: Optional[List[Tuple[str, str, str]]] = None,
    output_folder: Optional[str] = None
  ) -> Graph:
    """Generate a knowledge graph from input text or messages.
    
    Args:
        input_data: Text string or list of message dicts
        model: Name of OpenAI model to use
        api_key (str): OpenAI API key for making model calls
        chunk_size: Max size of text chunks to process
        context: Description of data context
        example_relations: Example relationship tuples
        node_labels: Valid node label strings
        edge_labels: Valid edge label strings
        ontology: Valid node-edge-node structure tuples
        output_folder: Path to save partial progress
        
    Returns:
        Generated knowledge graph
    """
    # TODO
    if chunk_size  or node_labels or edge_labels or ontology or example_relations or context:
      raise ValueError("chunk_size, node_labels, edge_labels, ontology, or example_relations are not supported parameters yet")
    
    # Process input data
    is_conversation = isinstance(input_data, list)
    if is_conversation:
      # Extract text from messages
      text_content = []
      for message in input_data:
        if not isinstance(message, dict) or 'role' not in message or 'content' not in message:
          raise ValueError("Messages must be dicts with 'role' and 'content' keys")
        if message['role'] in ['user', 'assistant']:
          text_content.append(f"{message['role']}: {message['content']}")
      
      # Join with newlines to preserve message boundaries
      processed_input = "\n".join(text_content)
    else:
      processed_input = input_data

    self.api_key = api_key
    self.model = model
    if self.api_key:
      self.lm = dspy.LM(model=self.model, api_key=self.api_key)
    else:
      self.lm = dspy.LM(model=self.model)
      
    self.dspy.configure(lm=self.lm)
    
    entities = get_entities(self.dspy, processed_input, is_conversation=is_conversation)
    relations = get_relations(self.dspy, processed_input, entities, is_conversation=is_conversation)
    
    graph = Graph(
      entities = entities,
      relations = relations,
      edges = {relation[1] for relation in relations}
    )
    
    if output_folder:
      os.makedirs(output_folder, exist_ok=True)
      output_path = os.path.join(output_folder, 'graph.json')
      
      graph_dict = {
        'entities': list(entities),
        'relations': list(relations),
        'edges': list(graph.edges)
      }
      
      with open(output_path, 'w') as f:
        json.dump(graph_dict, f, indent=2)
      
    return graph
    
  def cluster():
    pass
    
  def aggregate():
    pass
    



from src.kg_gen import KGGen
import os
from dotenv import load_dotenv

if __name__ == "__main__":
  # Load environment variables
  load_dotenv()
  
  # Example usage
  kg = KGGen()
  
  # Generate a simple graph
  text = "Linda is Josh's mother. Ben is Josh's brother. Andrew is Josh's father. Judy is Andrew's sister. Josh is Judy's nephew. Judy is Josh's aunt." # text option
  # text = [
  #   {"role": "user", "content": "Who is in Josh's family?"},
  #   {"role": "assistant", "content": "Josh's family includes his mother Linda, father Andrew, brother Ben, and aunt Judy who is Andrew's sister."}
  # ] # messages array
  graph = kg.generate(
    input_data=text,
    model="openai/gpt-4o",
    api_key=os.getenv("OPENAI_API_KEY")
  )
  print(graph)


from src.kg_gen import KGGen
import os
from dotenv import load_dotenv

if __name__ == "__main__":
  # Load environment variables
  load_dotenv()
  
  # Example usage
  kg = KGGen()
  
  # Generate a simple graph from conversation
  messages = [
    {"role": "user", "content": "What is the capital of France?"},
    {"role": "assistant", "content": "The capital of France is Paris."}
  ]
  
  graph = kg.generate(
    input_data=messages,
    model="openai/gpt-4o", 
    api_key=os.getenv("OPENAI_API_KEY")
  )
  print(graph)
